/*
	board.cc	Game Board Operations
	Copyright (c) 2000, 2001, 2003, 2004 Kriang Lerdsuwanakij
	email:		lerdsuwa@users.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "board.h"
#include "iter.h"
#include "hash.h"
#include <string.h>

#include "table.dat"

board_type board_empty = {
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
};

board_type board_begin = {
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0,-1, 1, 0, 0, 0, 
	 0, 0, 0, 1,-1, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
	 0, 0, 0, 0, 0, 0, 0, 0, 
};

// Flipable in 3 directions
#define FT3(a,b,c)	((1<<(a)) + (1<<(b)) + (1<<(c)))
#define FT3SE		FT3(DIR_E, DIR_SE, DIR_S)
#define FT3SW		FT3(DIR_W, DIR_SW, DIR_S)
#define FT3NE		FT3(DIR_E, DIR_NE, DIR_N)
#define FT3NW 		FT3(DIR_W, DIR_NW, DIR_N)

// Flipable in 5 directions
#define FT5(a,b,c,d,e)	((1<<(a)) + (1<<(b)) + (1<<(c)) + (1<<(d)) + (1<<(e)))
#define FT5S		FT5(DIR_E, DIR_SE, DIR_S, DIR_SW, DIR_W)
#define FT5E		FT5(DIR_N, DIR_NE, DIR_E, DIR_SE, DIR_S)
#define FT5W		FT5(DIR_N, DIR_NW, DIR_W, DIR_SW, DIR_S)
#define FT5N		FT5(DIR_E, DIR_NE, DIR_N, DIR_NW, DIR_W)

// Flipable in 8 directions
#define FT8		255

// Table of flipable directions in each position
int flip_table[64] = {
	FT3SE, FT3SE, FT5S,  FT5S,  FT5S,  FT5S,  FT3SW, FT3SW,
	FT3SE, FT3SE, FT5S,  FT5S,  FT5S,  FT5S,  FT3SW, FT3SW,
	FT5E,  FT5E,  FT8,   FT8,   FT8,   FT8,   FT5W,  FT5W,
	FT5E,  FT5E,  FT8,   FT8,   FT8,   FT8,   FT5W,  FT5W,
	FT5E,  FT5E,  FT8,   FT8,   FT8,   FT8,   FT5W,  FT5W,
	FT5E,  FT5E,  FT8,   FT8,   FT8,   FT8,   FT5W,  FT5W,
	FT3NE, FT3NE, FT5N,  FT5N,  FT5N,  FT5N,  FT3NW, FT3NW,
	FT3NE, FT3NE, FT5N,  FT5N,  FT5N,  FT5N,  FT3NW, FT3NW
};

		// No default initialization to speed up search
board_info::board_info()
{
}

board_info::board_info(const board_type *board_init)
{
	memcpy(board, board_init, sizeof(board_type));
	move = NUM_MOVE;
	board_iterator	iter;
	iter.init_pos();
	do {
		if (is_empty(this, iter.pos))
			move--;
	} while (iter.next());
	hash = get_hash(this);
}

bool operator==(const board_info &board1, const board_info &board2)
{
	if (memcmp(board1.board, board2.board, 64))
		return false;
	return true;
}

void	copy_board(board_info *dest, const board_info *src)
{
	memcpy(dest, src, sizeof(board_info));
}

int	board_diff_score(const board_info *board)
{
	int	score = 0;
	board_iterator	iter;
	iter.init_pos();
	do {
		score += board->board[iter.pos];
	} while (iter.next());
	return score;
}

int	board_black_score(const board_info *board)
{
	int	score = 0;
	board_iterator	iter;
	iter.init_pos();
	do {
		if (board->board[iter.pos] == BLACK)
			score++;
	} while (iter.next());
	return score;
}

int	board_white_score(const board_info *board)
{
	int	score = 0;
	board_iterator	iter;
	iter.init_pos();
	do {
		if (board->board[iter.pos] == WHITE)
			score++;
	} while (iter.next());
	return score;
}

bool	can_play_nocheck(const board_info *board, int color, int pos)
{
	board_dir_iterator	iter;
	for (int dir = DIR_START; dir <= DIR_END; ++dir) {

		if (flip_table[pos] & (1 << dir)) {
			iter.init_pos(pos);

			int	count = 0;
			while (iter.next_dir(dir)) {
				piece_type p = board->board[iter.pos];
				if (p == EMPTY) {
					break;
				}
				else if (p == color) {
					if (count) {
						return true;
					}
					else {
						break;
					}
				}
				else {
					count++;
				}
			}
		}
	}
	return false;
}

bool	can_play(const board_info *board, int color)
{
	board_iterator	iter;
	iter.init_pos();
	do {
		if (can_play(board, color, iter.pos))
			return true;
	} while (iter.next());
	return false;
}

bool	can_play(const board_info *board)
{
	return can_play(board, BLACK) || can_play(board, WHITE);
}

// Assume can_play(...) is true for color and pos
void	place_piece(board_info *board, int color, int pos)
{
	board->board[pos] = color;		// Place the new piece
	board->move++;
	board->hash ^= get_hash_piece(color, pos);

						// Flip pieces in all directions
	board_dir_iterator	iter;
	for (int dir = DIR_START; dir <= DIR_END; ++dir) {

		if (flip_table[pos] & (1 << dir)) {
			iter.init_pos(pos);

			int	count = 0;
			while (iter.next_dir(dir)) {
				piece_type p = board->board[iter.pos];
				if (p == EMPTY) {	// Not flanked
					break;
				}
				else if (p == color) {
					while (count) {	// Flanked
							// Flip
						iter.next_dir_nocheck(dir ^ 4);
						board->board[iter.pos] = color;
						board->hash ^= get_hash_flip(iter.pos);
						count--;
					}
					break;
				}
				else {
					count++;
				}
			}
		}
	}
}

bool	can_play_nocheck(const board_info *board, int color, int pos, int dir_count[NUM_DIR])
{
	bool	ret = false;
	board_dir_iterator	iter;
	for (int dir = DIR_START; dir <= DIR_END; ++dir) {

		if (flip_table[pos] & (1 << dir)) {
			iter.init_pos(pos);

			int	count = 0;
			dir_count[dir] = 0;

			while (iter.next_dir(dir)) {
				piece_type p = board->board[iter.pos];
				if (p == EMPTY) {
					break;
				}
				else if (p == color) {
					if (count) {
						ret = true;
						dir_count[dir] = count;
					}
					break;
				}
				else {
					count++;
				}
			}
		}
		else
			dir_count[dir] = 0;
	}
	return ret;
}

// Assume can_play(...) is true for color and pos
void	place_piece(board_info *board, int color, int pos, int dir_count[NUM_DIR])
{
	board->board[pos] = color;		// Place the new piece
	board->move++;
	board->hash ^= get_hash_piece(color, pos);

						// Flip pieces in all directions
	board_dir_iterator	iter;
	for (int dir = DIR_START; dir <= DIR_END; ++dir) {

		if (flip_table[pos] & (1 << dir)) {

			if (dir_count[dir]) {
				iter.init_pos(pos);

				int count = dir_count[dir];
				while (count) {
					iter.next_dir(dir);
					board->board[iter.pos] = color;
					board->hash ^= get_hash_flip(iter.pos);
					count--;
				}
			}
		}
	}
}
