/*
	board.h		Game Board Operations
	Copyright (c) 2000, 2001, 2002, 2003, 2004 Kriang Lerdsuwanakij
	email:		lerdsuwa@users.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef BOARD_H
#define BOARD_H

#include "config.h"

// Board squares
#define EMPTY	0
#define BLACK	1
#define WHITE	-1

// Total number of moves
#define NUM_MOVE	60

// Directions
#define NUM_DIR		8
#define DIR_START	0
#define DIR_END		7

// Impossible score to signal position evaluation
#define INVALID_SCORE	1000
#define MAX_SCORE	999
#define MIN_SCORE	-999

// Impossible position to signal a pass
#define POS_PASS 64
// Impossible position
#define POS_INVALID 65


// Board data types
typedef signed char piece_type;
typedef piece_type board_type[64];

extern board_type board_begin;

// x, y are 0-based.  Returns position in the 64-element array
inline int xy_to_pos(int x, int y)
{
	return y * 8 + x;
}

inline int pos_to_x(int pos)
{
	return pos % 8;
}

inline int pos_to_y(int pos)
{
	return pos / 8;
}

struct board_info {
	board_type	board;
	int		hash;
	char		move;
	board_info();
	board_info(const board_type *board_init);
};

bool operator==(const board_info &board1, const board_info &board2);

inline bool is_empty(const board_info *board, int x, int y)
{
	return board->board[xy_to_pos(x, y)] == EMPTY;
}

inline bool is_empty(const board_info *board, int pos)
{
	return board->board[pos] == EMPTY;
}

void	copy_board(board_info *dest, const board_info *src);

inline void copy_board(board_info &dest, const board_info &src)
{
	copy_board(&dest, &src);
}

inline void copy_board(board_info *dest, const board_type *src)
{
	board_info temp_board(src);
	copy_board(dest, &temp_board);
}

inline void copy_board(board_info &dest, const board_type *src)
{
	copy_board(&dest, src);
}

int	board_diff_score(const board_info *board);
int	board_black_score(const board_info *board);
int	board_white_score(const board_info *board);


inline void adjust_score(int &black_score, int &white_score)
{
	if (black_score + white_score < 64) {
		if (black_score > white_score)
			black_score = 64 - white_score;
		else if (black_score < white_score)
			white_score = 64 - black_score;
		else
			// FIXME: Get the scoring right
			;
	}
}

bool	can_play_nocheck(const board_info *board, int color, int pos);
inline bool	can_play(const board_info *board, int color, int pos)
{
					// This seems to be get the
					// fastest code
	if (!is_empty(board, pos))
		return false;
	return can_play_nocheck(board, color, pos);
}

bool	can_play_nocheck(const board_info *board, int color, int pos, int dir_count[NUM_DIR]);
inline bool	can_play(const board_info *board, int color, int pos, int dir_count[NUM_DIR])
{
					// This seems to be get the
					// fastest code
	if (!is_empty(board, pos))
		return false;
	return can_play_nocheck(board, color, pos, dir_count);
}

bool	can_play(const board_info *board, int color);
bool	can_play(const board_info *board);

void	place_piece(board_info *board, int color, int pos);
void	place_piece(board_info *board, int color, int pos, int dir_count[NUM_DIR]);

// Switch BLACK <-> WHITE
inline int	switch_color(int color) { return -color; }

inline int	get_num_move(const board_info *board) { return board->move; }

#endif /* BOARD_H */
