/*
	game.cc		Game States
	Copyright (c) 2004 Kriang Lerdsuwanakij
	email:		lerdsuwa@users.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"

#include <gtk/gtk.h>
#include <gnome.h>

#include <pthread.h>

#include <pwd.h>
#include <sys/types.h>
#include <time.h>

#include CXX__HEADER_exception
#include CXX__HEADER_stdexcept
#include CXX__HEADER_fstream
#include CXX__HEADER_sstream
#include CXX__HEADER_iomanip
#include <string>
#include <deque>
#include <vector>

USING_NAMESPACE_STD;

#include "game.h"
#include "opening.h"

#ifdef _
#  undef _
#endif

#ifdef N_
#  undef N_
#endif

#define _(x) (x)
#define N_(x) (x)

/*************************************************************************
	AI variables
*************************************************************************/

pthread_t	ai_thread;
pthread_attr_t	ai_thread_attr;
pthread_mutex_t	ai_lock = PTHREAD_MUTEX_INITIALIZER;
volatile int	ai_pos;			// Lock required for access
gint		ai_timeout_id;		// Lock required for access
bool		ai_tampered;

/*************************************************************************
	Game states
*************************************************************************/

board_info *board_ptr;
board_info *init_board_ptr;

int player;
bool wait_player;
bool game_play;
bool random_game;

// Opening information included
board_info	board_history[NUM_MOVE+1];
int		player_history[NUM_MOVE+1];
int		move_history[NUM_MOVE+1];
int		num_history;
int		min_num_history;
int		max_num_history;

deque<int>	move_queue;

string		home_dir;

vector<update_func> update_hook;

/*************************************************************************
	Configurations - Not front-end specific
*************************************************************************/

game_mode_type game_mode;

const char *game_mode_name[4] = {
	N_("None"), N_("Black"), N_("White"), N_("Both")
};

level_info_type level_info[NUM_LEVEL_INFO] = {
	{ 2, 6, 4 },
	{ 4, 8, 6 },
	{ 6, 12, 10 },
	{ 8, 16, 14 },
	{ 10, 20, 18 }
};

int computer_level = 2;

level_info_type current_level_info;

const char *level_name[NUM_LEVEL_INFO+1] = {
	N_("1"), N_("2"), N_("3"), N_("4"), N_("5"), N_("Custom")
};

int opening_var;
const char *opening_var_name[NUM_OPENING_VAR] = {
	N_("None"), N_("Very low"), N_("Low"), N_("Medium"), N_("High"), N_("Very high")
};

bool hint_move;
bool log_move;
bool animate_opening;
bool redo_ai_move;

int animate_delay;

string opening_name;
int opening_player;
deque<int> opening_move_queue;

const char *start_game_mode_name[NUM_START_GAME_MODE] = {
	N_("Initial board"), N_("Random board")
};
start_game_mode_type start_game_mode;
int start_random_game_pieces;

bool show_border;
bool show_toolbar;
bool show_pattern_evaluation;
bool show_game_history;
string theme_name;

/*************************************************************************
	Game mode functions
*************************************************************************/

game_mode_type get_game_mode()
{
	return game_mode;
}

void set_game_mode(game_mode_type game_mode_)
{
	game_mode = game_mode_;
}

const char *get_game_mode_string()
{
	switch (game_mode) {
		case COMPUTER_NONE:
			return _("None");
		case COMPUTER_BLACK:
			return _("Black");
		case COMPUTER_WHITE:
			return _("White");
		case COMPUTER_BOTH:
			return _("Both");
		default:
			return 0;
	}
}

bool is_computer_player(int color)
{
	if (color == BLACK) {
		if (get_game_mode() == COMPUTER_BLACK
		    || get_game_mode() == COMPUTER_BOTH)
			return true;
		return false;
	}
	else {
		if (get_game_mode() == COMPUTER_WHITE
		    || get_game_mode() == COMPUTER_BOTH)
			return true;
		return false;
	}
}

bool get_wait_player(int color)
{
	return !is_computer_player(color) && !move_queue.size();
}

bool get_game_play()
{
	return can_play(board_ptr);
}

/*************************************************************************
	File I/O functions
*************************************************************************/

void user_home_dir_init()
{
	struct passwd	*p;
		
					// Get passwd entry for current user
	p = getpwuid(getuid());
	if (p == NULL) {
		char *env = getenv("HOME");
		if (env == NULL) {
					// Give up
			throw runtime_error(_("error obtaining the user home directory"));
		}
	
		home_dir = env;
	}
	else {
			
					// Store home directory
		home_dir = p->pw_dir;
	}
}

void log_history()
{
					// No need to do i18n here
	if (!log_move)
		return;

					// Random game not supported yet
	if (start_game_mode != START_GAME_INITIAL)
  		return;

	string	file = home_dir;
	file += "/grhino.log";
	ofstream of(file.c_str(), ios::app);

	of << "         0 ";

	int	black_score = board_black_score(&board_history[max_num_history-1]);
	int	white_score = board_white_score(&board_history[max_num_history-1]);
	of << "e ";
	adjust_score(black_score, white_score);

	of << ((!ai_tampered && is_computer_player(BLACK)) ? "grhino-B " : "player-B ")
	   << setw(2) << black_score << " ( 30   0   0) ";
	of << ((!ai_tampered && is_computer_player(WHITE)) ? "grhino-W " : "player-W ")
	   << setw(2) << white_score << " ( 30   0   0) ";

	for (int i = 1; i < max_num_history; ++i) {
		if (player_history[i-1] == BLACK)
			of << '+';
		else
			of << '-';
		of << static_cast<char>('1' + pos_to_x(move_history[i-1]))
		   << static_cast<char>('1' + pos_to_y(move_history[i-1]));
	}

	of << " +0" << endl;
}

/*************************************************************************
	State update
*************************************************************************/

void update_state(update_state_type u)
{
	for (size_t i = 0; i < update_hook.size(); ++i)
		update_hook[i](u);
}

void add_update_hook(update_func f)
{
	update_hook.push_back(f);
}

void remove_update_hook(update_func f)
{
	size_t i = 0;
	size_t size = update_hook.size();
	for ( ; i < size; ++i) {
		if (update_hook[i] == f)
			break;
	}
			// Not found
	if (i == size)
		return;

	if (i != size-1)
		update_hook[i] = update_hook[size-1];
	update_hook.pop_back();
}

/*************************************************************************
	Utility functions
*************************************************************************/

const char *find_opening_name()
{
	for (int i = 0; i < get_num_opening(); ++i) {
				// Wrong number of moves
		if (board_ptr->move != get_opening_num_move(i))
			continue;

		int j, k;

				// Straightforward compare
		for (j = 0, k = 0; j < 64; ++j, ++k) {
			if (board_ptr->board[j] != (*get_opening_board(i))[k])
				break;
		}
		if (j == 64)
			return get_opening_name(i);

				// Transpose
		for (j = 0, k = 0; j < 64; ++j, k = (k>=56 ? (k%8)+1: k+8)) {
			if (board_ptr->board[j] != (*get_opening_board(i))[k])
				break;
		}
		if (j == 64)
			return get_opening_name(i);

				// Rotate
		for (j = 0, k = 63; j < 64; ++j, --k) {
			if (board_ptr->board[j] != (*get_opening_board(i))[k])
				break;
		}
		if (j == 64)
			return get_opening_name(i);

				// Rotate and transpose
		for (j = 0, k = 63; j < 64; ++j, k = (k<8 ? (k%8)+55: k-8)) {
			if (board_ptr->board[j] != (*get_opening_board(i))[k])
				break;
		}
		if (j == 64)
			return get_opening_name(i);
	}
	return 0;
}

