/*
	log_proc.h	Log file processing
	Copyright (c) 2001, 2002, 2004 Kriang Lerdsuwanakij
	email:		lerdsuwa@users.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"

#include <string>
#include CXX__HEADER_fstream
#include CXX__HEADER_iomanip

#include "board.h"
#include "order.h"
#include "proginfo.h"

USING_NAMESPACE_STD;

struct game_log {
	int move_queue[60];
	int move_queue_color[60];
	int num_move_queue;
	board_info board;
	int black_score;
	int white_score;

	game_log(int bs, int ws): 
		num_move_queue(0), board(),
		black_score(bs), white_score(ws) {}
	void copy_board(board_type *b) { ::copy_board(&board, b); }
	void copy_board(board_info *b) { ::copy_board(&board, b); }
	void push_move(int pos, int color) {
		move_queue[num_move_queue] = pos;
		move_queue_color[num_move_queue] = color;
		num_move_queue++;
	}
	void reset_move() { num_move_queue = 0; }
};

#define VAL_BLANK	-1
#define VAL_WHITE	-2
#define VAL_BLACK	-3

int	from_ascii(const string &s, int i)
{
	if (s[i] == ' ') {
		if (s[i+1] == ' ')
			return VAL_BLANK;
		else
			return s[i+1]-'0';
	}
	else if (s[i] == '(' && s[i+1] == ')')
		return VAL_WHITE;
	else if (s[i] == '#' && s[i+1] == '#')
		return VAL_BLACK;
	else {
		return (s[i]-'0')*10 + s[i+1]-'0';
	}
}

template <class T> void	process_line_IOS(const string &s, T &t, bool need_end)
{
	int	black_score = from_ascii(s, 22);
	int	white_score = from_ascii(s, 48);

	if (s[11] != 'e' && need_end)	// Game not properly ended
		return;

				// Old IOS format
	if (s[65] == '+') {
		int index = 65;

		game_log game(black_score, white_score);
		game.copy_board(&board_begin);

		do {
			int player = (s[index] == '+') ? BLACK : WHITE;
			int ipos = from_ascii(s, index+1);
			int pos = xy_to_pos(ipos/10 -1, ipos%10 -1);
			game.push_move(pos, player);
			index += 3;
		} while (s[index] == '+' || s[index] == '-');

		process_game(t, game);
	}
				// Old IOS format, resigned game
	else if (s[66] == '+')
		;
				// Old IOS format with 'K'
	else if (s[65] == 'K' && s[70] == '+') {
		int index = 70;

		game_log game(black_score, white_score);
		game.copy_board(&board_begin);

		do {
			int player = (s[index] == '+') ? BLACK : WHITE;
			int ipos = from_ascii(s, index+1);
			int pos = xy_to_pos(ipos/10 -1, ipos%10 -1);
			game.push_move(pos, player);
			index += 3;
		} while (s[index] == '+' || s[index] == '-');

		process_game(t, game);
	}
				// New IOS format
	else if (s[65] == 'K' && s[72] == ' ') {
		board_info board;
		board_type init_board;

		int num_move_queue = 0;
		int move_queue[60];
		int index = 73;

				// New IOS format support random games
				// Setup the initial board
		for (int pos = 0; pos < 64; ++pos, index += 2) {
			int val = from_ascii(s, index);
			switch (val) {
				case VAL_WHITE:
					init_board[pos] = WHITE;
					break;
				case VAL_BLACK:
					init_board[pos] = BLACK;
					break;
				case VAL_BLANK:
					init_board[pos] = EMPTY;
					break;
				default:
					init_board[pos] = EMPTY;
					move_queue[val-1] = pos;
					num_move_queue++;
			}
		}

		board_info temp_board(&init_board);
		if (temp_board.move == 0) {
			if (init_board[D4] != WHITE)
				return;
			if (init_board[D5] != BLACK)
				return;
			if (init_board[E4] != BLACK)
				return;
			if (init_board[E5] != WHITE)
				return;
		}

				// Assume that black moves first
		copy_board(board, temp_board);

				// Scores will be filled later
		game_log game(0, 0);

		int init_player = BLACK;
		int player = init_player;
		int i;
		for (i = 0; i < num_move_queue; ++i) {
				// Fail, wrong assumption about to get
				// the first move
			if (!can_play(&board, player, move_queue[i])) {
				break;
			}

			game.push_move(move_queue[i], player);

			place_piece(&board, player, move_queue[i]);
			if (can_play(&board, switch_color(player)))
				player = switch_color(player);
		}
		if (i == num_move_queue) {
			black_score = board_black_score(&board);
			white_score = board_white_score(&board);
			adjust_score(black_score, white_score);
		}
		else {
				// Assume that white moves first

			game.reset_move();
			copy_board(board, temp_board);

			init_player = WHITE;
			player = init_player;
			for (int i = 0; i < num_move_queue; ++i) {
				if (!can_play(&board, player, move_queue[i]))
					abort();

				game.push_move(move_queue[i], player);

				place_piece(&board, player, move_queue[i]);
				if (can_play(&board, switch_color(player)))
					player = switch_color(player);
			}
			black_score = board_black_score(&board);
			white_score = board_white_score(&board);
			adjust_score(black_score, white_score);
		}

				// Now collect pattern info
		game.copy_board(&temp_board);
		game.black_score = black_score;
		game.white_score = white_score;

		process_game(t, game);
	}
}

// Input i = position after "(;"
// Return position after ";)" or -1 upon error
template <class T> int	process_game_GGS(const string &s, int i, T &t, bool need_end)
{
	board_type init_board;

	int size = s.size();

				// Scores will be filled later
 	game_log game(0, 0);

	try {
		while (i < size) {
			char c1 = s.at(i);
			char c2 = s.at(i+1);

			if (c1 == ';' && c2 == ')') {
				process_game(t, game);
				return i+2;
			}
	
			char c3 = s.at(i+2);		// At least 3 chars needed

							// Sorted according to
							// order of appearance
			if (c1 == 'G' && c2 == 'M' && c3 == '[') {
				i += 3;
				if (s.compare(i, 8, "Othello]"))
					return -1;

				while (s.at(i) != ']')
					i++;
				i++;
			}
			else if (c1 == 'T' && c2 == 'Y' && c3 == '[') {
				i += 3;
				char c;
				bool found8 = false;
				while ((c = s.at(i)) != ']') {
					if (c == 'a')		// Anti game
						return -1;
					else if (c == '8') {
						if (found8)	// 88 game
							return -1;
						else
							found8 = true;
						i++;
					}
					else if (c == 'r')	// Random game
						break;		// skip remaining
					else if (c >= '0' && c <= '9')	// Not 8x8 board
						return -1;
					else
						i++;
				}

				while (s.at(i) != ']')
					i++;
				i++;
			}
			else if (c1 == 'R' && c2 == 'E' && c3 == '[') {
				i += 3;

							// Read score difference
				int diff_sign = (s.at(i) == '+') ? 1: -1;
				i++;
				int diff = 0;
				while (s.at(i) >= '0' && s.at(i) <= '9') {
					diff = diff*10 + (s.at(i)-'0');
					i++;
				}
				diff = diff * diff_sign;

							// Check game result
				while (s.at(i) != ']') {
					if (s.at(i) == 't' && need_end)	// Timeout
						return -1;
					if (s.at(i) == 'r' && need_end)	// Resign
						return -1;
					if (s.at(i) == 's' && need_end)	// Mutual score
						return -1;
					i++;
				}

				while (s.at(i) != ']')
					i++;
				i++;

							// Correct when diff == 0
				game.black_score = diff > 0 ? diff : 0;
				game.white_score = diff < 0 ? diff : 0;
			}
			else if (c1 == 'B' && c2 == 'O' && c3 == '[') {
				i += 3;
				if (s.at(i) != '8' || s.at(i+1) != ' ')	// Bad board size
					return -1;
				i += 2;

							// GGS format support random games
							// Setup the initial board
				for (int pos = 0; pos < 64; ++pos) {

					while (s.at(i) == ' ')
						i++;

					switch (s.at(i)) {
						case 'O':
							init_board[pos] = WHITE;
							break;
						case '*':
							init_board[pos] = BLACK;
							break;
						case '-':
							init_board[pos] = EMPTY;
							break;
						default:
							return -1;	// Unknown character
					}
					i++;	
				}
				game.copy_board(&init_board);
								// No need to process color
								// to move.  It's handled by
								// B[...] or W[...]

				while (s.at(i) != ']')
					i++;
				i++;
			}
			else if (c1 == 'B' && c2 == '[') {
				i += 2;
				if (s.at(i) == 'P' || s.at(i) == 'p')
					;	// Pass
				else if (s.at(i) >= 'A' && s.at(i) <= 'H') {
					int pos = xy_to_pos(s.at(i)-'A', s.at(i+1)-'1');
					game.push_move(pos, BLACK);
				}
				else if (s.at(i) >= 'a' && s.at(i) <= 'h') {
					int pos = xy_to_pos(s.at(i)-'a', s.at(i+1)-'1');
					game.push_move(pos, BLACK);
				}
				else
					return -1;

				while (s.at(i) != ']')
					i++;
				i++;
			}
			else if (c1 == 'W' && c2 == '[') {
				i += 2;
				if (s.at(i) == 'P' || s.at(i) == 'p')
					;	// Pass
				else if (s.at(i) >= 'A' && s.at(i) <= 'H') {
					int pos = xy_to_pos(s.at(i)-'A', s.at(i+1)-'1');
					game.push_move(pos, WHITE);
				}
				else if (s.at(i) >= 'a' && s.at(i) <= 'h') {
					int pos = xy_to_pos(s.at(i)-'a', s.at(i+1)-'1');
					game.push_move(pos, WHITE);
				}
				else
					return -1;

				while (s.at(i) != ']')
					i++;
				i++;
			}
			else {
				if (isspace(c1))
					i++;
				else {
							// Skip the remaining property
							// also ignore garbage char after
							// ']' like W[a2]//5.53]
					while (s.at(i) != '[' && s.at(i) != ']')
						i++;

					if (s.at(i) == '[') {	// Skip property
						i++;		// Move after '['
						while (s.at(i) != ']')
							i++;
						i++;		// Move after ']'
					}
					else
						i++;		// Move after ']'
				}
			}
		}
							// Premature end-of-line
		return -1;
	}
	catch (out_of_range &) {
		return -1;
	}
}

template <class T> void	process_line_GGS(const string &s, T &t, bool need_end)
{
	int i = 0;
	int size = s.size();

	while (i+1 < size) {
		if (s[i] == '(' && s[i+1] == ';') {
			i = process_game_GGS(s, i+2, t, need_end);
			if (i == -1)		// Erroneous line
				return;
		}
		else
			i++;
	}
}

template <class T> void	process_line(const string &s, T &t, bool need_end)
{
	if ((s[0] == '(' && s[1] == ';')	// Old GGS
	    || (s[2] == '(' && s[3] == ';'))	// New GGS supporting synchro rand
		process_line_GGS(s, t, need_end);
	else
		process_line_IOS(s, t, need_end);
}

template <class T> void	process_file(const char *f, T &t, bool need_end = true)
{
	cout << "Processing " << f << endl;
	ifstream ifs(f);
	if (!ifs) {
		cerr << prog_name << ": cannot open file " << f << endl;
		throw 1;
	}

	int line = 0;
	for ( ; ; ) {
		if (line % 100 == 99) {
			cout << '.';
			cout.flush();
		}
		line++;

		string buffer;
		getline(ifs, buffer);
					// End of file
		if (!ifs)
			break;

		try {
			process_line(buffer, t, need_end);
		}
		catch (...) {
			cerr << '\n' << prog_name << ": fail at line " << line << endl;
		}
	}
	cout << endl;
}
